#include <stdlib.h>
#include "spawn.h"
#include <fstream.h>

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"

#define UNASSIGNED 0xffff


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::ResetLimits()
{
	//If _limits negative, or _limits unset, set to 300, 300, x, x, 2, 2.
	if(currping_limit < 0)	currping_limit = 300;
	if(avgping_limit < 0)	avgping_limit = 300;
	if(stcploss_limit < 0)	stcploss_limit = 2;
	if(ctsploss_limit < 0)	ctsploss_limit = 2;
	if(stcweapon_limit < 0)	stcweapon_limit = 2;
}
void botInfo::LoadINI()
{
	ifstream file("lagbot.ini");
	char line[256];
	while (file.getline(line, 256))
	{
		if (CMPSTART("WarnMessage=", line))
		{
			WarnMessage = "*warn ";
			WarnMessage += &line[12];
		}
		else if (CMPSTART("LockMessage=", line))
		{
			LockMessage = "*warn ";
			LockMessage += &line[12];
		}

		else if (CMPSTART("current=", line))
			currping_limit =  atoi(&(line[8]));
		else if (CMPSTART("average=", line))
			avgping_limit =  atoi(&(line[8]));
		else if (CMPSTART("s2cploss=", line))
			stcploss_limit =  atoi(&(line[9]));
		else if (CMPSTART("c2sploss=", line))
			ctsploss_limit =  atoi(&(line[9]));
		else if (CMPSTART("s2cweapons=", line))
			stcweapon_limit =  atoi(&(line[11]));

		else if (CMPSTART("display=", line))
			showlag_status =  atoi(&(line[8]));
		else if (CMPSTART("deal=", line))
			laggerwarn_status =  atoi(&(line[5]));
		else if (CMPSTART("maxcheckrate=",line))
			lagtimelimit = atoi(&line[13]);
	}
	file.close();
	ResetLimits();
}


void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];
			_listnode <Player> *parse = playerlist->head;
			while (parse)
			{
				Player *p = parse->item;
				int time = get_tag(p, 0);
				set_tag(p, 0, --time);
				parse = parse->next;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
			bool biller_online = *(bool*)&event.p[2];
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			killTags();
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerWeapon:
		{
			Player *p = (Player*)event.p[0];
			weaponInfo wi;
			wi.n = *(Uint16*)&event.p[1];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
			Uint16 bounty = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
			int prize = *(int *)&event.p[1];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldship = (Uint16)event.p[1];
			Uint16 oldteam = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			killTags(p);
		}
		break;
//////// Selfish ////////
	case EVENT_SelfShipReset:
		{
		}
		break;
	case EVENT_SelfPrize:
		{
			Uint16 prize = (Uint16)event.p[0];
			Uint16 count = (Uint16)event.p[1];
		}
		break;
	case EVENT_SelfUFO:
		{
		}
		break;
	case EVENT_PositionHook:
		{
		}
		break;
//////// Bricks ////////
	case EVENT_BrickDropped:
		{
			int x1 = *(int*)&event.p[0];
			int y1 = *(int*)&event.p[1];
			int x2 = *(int*)&event.p[2];
			int y2 = *(int*)&event.p[3];
			int team = *(int*)&event.p[4];
		}
		break;
//////// LVZ ////////
	case EVENT_ObjectToggled:
		{
			objectInfo obj;
			obj.n = (Uint16)event.p[0];
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					String lagreturns = msg;
					String isitping = lagreturns.split(' ');
					if(isitping != "PING")					//Prove *lag return
						break;
					

					//Destruction of returned msg
					String junk = lagreturns.split(':');	//junk now set to "Current"
					currping_s = lagreturns.split(' ');		//currping now set to "####"
					junk = lagreturns.split(' ');			//junk now set to "ms"      what we have left is "Average:30 ms  Low:0 ms  High:790 ms  S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(':');			//junk now set to "Average"
					avgping_s = lagreturns.split(' ');		//avgping set to "####"     what we have left is "ms  Low:0 ms  High:790 ms  S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(' ');			//junk now set to "ms"
					junk = lagreturns.split(':');			//junk now set to "Low"
					lowping_s = lagreturns.split(' ');		//lowping set to "####"     what we have left is "ms High:790 ms  S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(' ');			//junk now set to "ms"      what we have left is "High:790 ms  S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(':');			//junk now set to "High"
					highping_s = lagreturns.split(' ');		//highping set to "####"    what we have left is "ms  S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(' ');			//junk now set to "ms"      what we have left is "S2C: 0.0%  C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(':');			//junk now set to "S2C"
					stcploss_s = lagreturns.split('%');		//stcploss set to "####"    what we have left is "C2S: 0.0%  S2CWeapons: 0.0%"
					junk = lagreturns.split(':');			//junk now set to "C2S"
					ctsploss_s = lagreturns.split('%');		//stcploss set to "####"    what we have left is "S2CWeapons: 0.0%"
					junk = lagreturns.split(':');			//junk now set to "S2CWeapons"
					stcweapon_s = lagreturns.split('%');	//stcploss set to "####"    what we have left is "NOTHING   WOOT"


					//Piece response
					/*
					responseA  = PlayerReqLagCheck;
					responseA += " requested a lag check for ";
					responseA += PlayerLagCheck;
					if(eavesdropper->name)
						sendPrivate(eavesdropper,responseA);
					*/
					
					if (showlag_status == 1)
					{
						
						//responseA  = "*arena ";
						response   = "*arena C: ";
						
						
						response  = "*arena [";
						response += PlayerLagCheck;
						response += "]  C: ";
					}
					else if (showlag_status == 2)
					{
						response  = "[";
						response += PlayerLagCheck;
						response += "]  C: ";
					}
					else
						break;
					response += currping_s;
					response += "  Avg: ";
					response += avgping_s;
					response += "  L: ";
					response += lowping_s;
					response += "  H: ";
					response += highping_s;
					response += "  S2C: ";
					response += stcploss_s;
					response += "  C2S: ";
					response += ctsploss_s;
					response += "  S2CW: ";
					response += stcweapon_s;

					if (showlag_status == 1)
					{
						sendPublic(response);
					}
					else if (showlag_status == 2)
					{
						_listnode <Player> *parse = playerlist->head;
						while (parse)
						{
							Player *p = parse->item;
							String PMLagTo = p->name;
							if (PMLagTo == PlayerReqLagCheck)
							{
								sendPrivate(p, response);
							}
							parse = parse->next;
						}
					}


					//Strings to ints
					currping_n  = atoi(currping_s);
					avgping_n   = atoi(avgping_s);
					lowping_n   = atoi(lowping_s);
					highping_n  = atoi(highping_s);
					stcploss_n  = atoi(stcploss_s);
					ctsploss_n  = atoi(ctsploss_s);
					stcweapon_n = atoi(stcweapon_s);

					ResetLimits();

					//Compare numbers to limits
					if(laggerwarn_status)
					{
						if(currping_n >= currping_limit)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += currping_limit;
							response += "+ PING Current";
							sendPublic(response);
							warn_status = true;
						}
						if(avgping_n >= avgping_limit)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += avgping_limit;
							response += "+ PING Average";
							sendPublic(response);
							warn_status = true;
						}
						if(stcploss_n >= stcploss_limit)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += stcploss_limit;
							response += "%+ S2CPloss";
							sendPublic(response);
							warn_status = true;
						}
						if(ctsploss_n >= ctsploss_limit)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += ctsploss_limit;
							response += "%+ C2SPloss";
							sendPublic(response);
							warn_status = true;
						}
						if(stcweapon_n >= stcweapon_limit)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += stcweapon_limit;
							response += "%+ S2CWeapons";
							sendPublic(response);
							warn_status = true;
						}

						if(stcploss_n < -2)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += "negative S2CPloss";
							sendPublic(response);
							warn_status = true;
						}
						if(ctsploss_n < -2)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += "negative C2SPloss";
							sendPublic(response);
							warn_status = true;
						}
						if(stcweapon_n < -2)
						{
							response  = "*arena";
							response += PlayerLagCheck;
							response += " is lagging with ";
							response += "negative S2CWeapons ploss";
							sendPublic(response);
							warn_status = true;
						}
					}
					if(  (laggerwarn_status == 2 || laggerwarn_status == 3)  && warn_status == true)
					{
						_listnode <Player> *parse = playerlist->head;
						while (parse)
						{
							Player *p = parse->item;
							String ReturnLag = p->name;
							if (ReturnLag == PlayerLagCheck)
							{
								if(laggerwarn_status == 2)
									sendPrivate(p, WarnMessage);
								else if(laggerwarn_status == 3)
								{
									sendPrivate(p, LockMessage);
									if(p->ship != SHIP_Spectator)
										sendPrivate(p, "*spec");//single *spec to lock
								}
							}
							parse = parse->next;
						}
					}
					warn_status = false;
					PlayerReqLagCheck = "";
					PlayerLagCheck= "";
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major	= HIWORD(*(int*)&event.p[0]);
			int minor	= LOWORD(*(int*)&event.p[0]);
			callback	= (CALL_COMMAND)event.p[1];
			playerlist	= (CALL_PLIST)event.p[2];
			flaglist	= (CALL_FLIST)event.p[3];
			map			= (CALL_MAP)event.p[4];
			bricklist	= (CALL_BLIST)event.p[5];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
			{
				tell(makeEcho("DLL plugin connected."));
				LoadINI();
			}

		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Tag emulation ////////

int botInfo::get_tag(Player *p, int index)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
			return tag->data;

		parse = parse->next;
	}

	return 0;
}

void botInfo::set_tag(Player *p, int index, int data)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
		{
			tag->data = data;
			return;
		}

		parse = parse->next;
	}

	tag = new PlayerTag;
	tag->p = p;
	tag->index = index;
	tag->data = data;
	taglist.append(tag);
}

void botInfo::killTags(Player *p)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;
		parse = parse->next;

		if (tag->p != p) continue;

		taglist.kill(tag);
	}
}

void botInfo::killTags()
{
	taglist.clear();
}


//////// LVZ Object Toggling ////////

void botInfo::clear_objects()
{
	num_objects = 0;
}

void botInfo::object_target(Player *p)
{
	object_dest = p;
}

void botInfo::toggle_objects()
{
	Player *p = object_dest;

	if (!p)	tell(makeToggleObjects(UNASSIGNED, (Uint16 *)object_array, num_objects));
	else	tell(makeToggleObjects(p->ident, (Uint16 *)object_array, num_objects));

	num_objects = 0;
}

void botInfo::queue_enable(int id)
{
	if (num_objects == MAX_OBJECTS)
		toggle_objects();

	object_array[num_objects].id = id;
	object_array[num_objects].disabled = false;
	++num_objects;
}

void botInfo::queue_disable(int id)
{
	if (num_objects == MAX_OBJECTS)
		toggle_objects();

	object_array[num_objects].id = id;
	object_array[num_objects].disabled = true;
	++num_objects;
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
